/*  unique filter -- reads file of strings from stdin and removes dupes  
		 
    Copyright (C) 1985,1987,1991,1994,1996,2003,2004  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

	Windows:     Compile with Visual Studio 5 or later or gcc 2.9.5 or later
	should work with just about any c compiler.
	cl unique.c
	windows or unix using gcc:
	gcc unique.c -o unique.exe
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#else
#define strnicmp strncasecmp
#define stricmp strcasecmp
#endif

#define CPMEOF 26
#define STDERR 2
#define LEN 65536
char str[LEN],pstr[LEN],buf[LEN+128]; /* each record may be up to 8KB */


void usage() {

	fprintf(stderr,"Unique Filter V2.00 %s\n",__DATE__);
	fprintf(stderr,"Copyright 1985-2004, Norm London, London Computing.\nnorm_london@londoncomputing.com\n\n");
	fprintf(stderr,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(stderr,"it under the terms of the GNU General Public License as published by\n");
	fprintf(stderr,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(stderr,"(at your option) any later version.\n\n");
	fprintf(stderr,"This program is distributed in the hope that it will be useful,\n");
	fprintf(stderr,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(stderr,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(stderr,"GNU General Public License for more details.\n\n");
	fprintf(stderr,"You should have received a copy of the GNU General Public License\n");
	fprintf(stderr,"along with this program; if not, write to the Free Software\n");
	fprintf(stderr,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
	fprintf(stderr,"Please press return to continue:");
	getchar();

	fprintf(stderr,"Usage is:\nUNIQUE [-?] [-Q] [-U] {-N|[-X[1]]-K}  [-S fldstart fldlen] [-A string] [[fldstart fldlen] ...] [<infile] [>outfile]\n");
	fprintf(stderr,"Extracts first occurrence of data records.  Data must be sorted.\n\n");
	fprintf(stderr,"You can define up to 31 fields as pairs of fldstart fldlen.\n");
	fprintf(stderr,"fldstart is starting postition relative to 1.\n");
	fprintf(stderr,"fldlen   is the length of the field in bytes.\n");
	fprintf(stderr,"Enter -U to ignore case during compare.\n");
	fprintf(stderr,"      -X to return all instances of duplicated records only.\n");
	fprintf(stderr,"      -X1 to return 1st instance of duplicated records only.\n");
	fprintf(stderr,"      -N to return only records that are NOT duplicated.\n");
	fprintf(stderr,"      -Q for Quiet Mode.  Do not print out instructions.\n");
	fprintf(stderr,"      -K Show count for Unique records.\n");
	fprintf(stderr,"      -A Concatenate the string to the end of the record\n");
	fprintf(stderr,"      -C Concatenate the field at fldstart of fldlen\n");
	fprintf(stderr,"         -C and -A are implemented in order on the command line.\n");
	fprintf(stderr,"      -? for help to be displayed.\n\n");
	fprintf(stderr,"ERRLEVEL set to\n\t0 if no duplicated records found,\n");
	fprintf(stderr,"\t1 if any found.\n");
	fprintf(stderr,"\t4 if -? is specified.\n");                       
}

// output string + suffixes if any specified
putstring(char *s,int argc,char *argv[]) {
	static int firsttime = 1;
	char buf[LEN];
	int i,start,len;
	fputs(s,stdout);
	for (i=1;i < argc; ++i) {
		if (argv[i][0] == '-') {
			switch(toupper(argv[i][1])) {
			case 'A':
				fputs(argv[++i],stdout);
				break;
			case 'C':
				start=atoi(argv[++i]);
				len=atoi(argv[++i]);
				if (len>0) {
					strncpy(buf,s+start,len);
					buf[len] = 0;
				}
				else
					strcpy(buf,s+start);
				fputs(buf,stdout);
				break;
			}
		}
	}
	fputs("\n",stdout);
}


int main(int argc,char *argv[]) { 
	char uflag,*s,*ps,*t, showcount = 0;

	int x1flag=0,quiet=0,nodupe=0,firsttime,xflag=0, start, len;
	int e1,b[32],e[32],l,pl,kt,k,new_value=0,dupe=0,help=0;
	long inkt,otkt, dupekt;
	static char badarg[]=" -- Invalid!\n";
	char *buf;

	
	/* look for flags, then get series of field defs up to 32 */
	for (l=1;l<argc && argv[l][0] == '-';++l) {
		switch (toupper(argv[l][1])) {
		case '?':
			help = 1;
			break;
		case 'K':
			showcount = 1;
			break;
		case 'N':
			nodupe = 1;
			xflag = 0;
			break;
		case 'Q':
			quiet = 1;
			break;
		case 'X':
			xflag = 1;
			nodupe = 0;
			if (argv[l][2] == '1')
				x1flag = 1;
			break;
		case 'U':
			uflag = 1;
			break;
		case 'A':
			if (++l >= argc) {
				fprintf(stderr,"-A needs to be followed by a string.\n");
				return 4;
			}
			break;
		case 'C':
			if (l > argc-1) {
				fprintf(stderr,"-C needs to be followed by a starting position and length.\n");
				return 4;
			}
			start = atoi(argv[++l]);
			len = atoi(argv[++l]);
			if (start < 1) {
				fprintf(stderr,"-C start, start must be >= 1\n");
				return(4);
			}
			if (len < 0) {
				fprintf(stderr,"-C start len, len must be >= 0\n");
				return(4);
			}
			break;
		default:
			fprintf(stderr,"%s%s",argv[l],badarg);
			break;
		}
	}
																															  
	if (argc < 2 || help) {
		usage();
		if (help)
			exit(4);
		else
			fprintf(stderr,"Type ^Z<RETURN> at end of input.\n\n");

	}

	/* get up to 32 field defs */
	for (kt=0;kt < 32 && l < argc;++kt) {
		if (argc > l)
			b[kt] = atoi(argv[l++]) - 1;
		if (argc > l)
			e[kt] = atoi(argv[l++]);
		if (b[kt] < 0)
			b[kt] = 0;
		if (e[kt] <= 0)
			e[kt] = 0;
	}
	str[0]=0;
	pstr[0]=0;
	s = str;
	ps= pstr;
	firsttime = 1;
	pl = 0;
	inkt = 0;
	otkt = 0;
	while (fgets(s,LEN,stdin) != NULL) {
		// Read a new string and compare to previous
		// set e1 to 1 if different, 0 if the same
#ifdef DEBUG
		printf("s = %s, ps = %s, new_value = %d\n",s,ps,new_value);
#endif          
		++inkt;
		l = strlen(s);
		s[--l] = 0;  //Remove trailing linefeed
		if (kt) {
			 /* if any arguments present */
			for (k = 0, e1 = 0; e1 == 0 && k < kt; ++k) {
				 /* if both substrings longer or equal to ln, do substr compare */
				if ((l+b[k]) >= e[k] && (pl+b[k]) >= e[k])
					if (uflag)
						e1=strnicmp(s+b[k],ps+b[k],e[k]);
					else
						e1 = strncmp(s+b[k],ps+b[k],e[k]);
				    /* if both strings of same length and length less than ln, do str compare */
				else if (l == pl)
					if (uflag)
						e1 = stricmp(s+b[k],ps+b[k]);
					else
						e1 = strcmp(s+b[k],ps+b[k]);
				    /* otherwise , not equal */
				else
					e1 = 1;
			}
		}
		else  /* compare on whole string */
			if (l == pl)
				if (uflag)
					e1 = stricmp(s,ps);
				else
					e1 = strcmp(s,ps);
			else
				e1 = 1;

		if (e1) {
			/* data is unique, set flag for -x usage */
			new_value = 1;
		}
		else {
			dupe = 1;
			++dupekt;
		}

		if (xflag)
			if (e1)
				e1 = 0;
			else
				e1 = 1;
			/* reverse meaning for -x */

#ifdef DEBUG
		printf("new_value = %d, e1 = %d\n",new_value,e1);
#endif

		if (e1) {
			/* Current record is not the same as previous record if xflag = 0 */
			/* Current record is     the same as previous record if xflag = 1 */
			if (nodupe) {
				/* Current record should be different */
				if (new_value) {
					new_value = 0;
					if (dupe)
						dupe = 0;
					else {
						/* if this is the first time, then output empty line*/
						if (firsttime)
							firsttime = 0;
						else {
							putstring(ps,argc,argv);
							otkt++;
						}
					}
				}
			}                                               
			else if (xflag) {
				if (new_value) {
					new_value = 0;
					putstring(ps,argc,argv);
					otkt++;
				}
			}
			if (x1flag == 0 && nodupe == 0) {
				if (showcount) {
					// write data when record changes, not when it first appears
					
					if (firsttime)
						firsttime = 0;
					else {
						sprintf(buf,"%12ld %s",dupekt,ps);
						putstring(buf,argc,argv);
					}
					dupekt = 1;
				}
				else 						
					putstring(s,argc,argv);
				otkt++;
			}
		}
		/* swap pointers to s, ps so that ps points to last record read */
		/* and copying is not necessary */
		t = s;
		s = ps;
		ps = t;
		pl = l;
	}
	if (nodupe) { /* If last record was not a dupe, write it out */
		/* Current record should be different */
		if (!dupe) {
			/* if this is the first time, then output empty line*/
			putstring(ps,argc,argv);
			otkt++;
		}
	}       
	else {
		if (showcount) {
			sprintf(buf,"%12ld %s\n",dupekt,ps);
			putstring(buf,argc,argv);
			otkt++;
		} 
	}
	if (!quiet)
		fprintf(stderr,"%ld records read.  %ld records written.\n",inkt,otkt);
	/* set return code */
	if (xflag) 
		if (otkt)
			k = 1;  /* found at least one duplicated record */
		else
			k = 0;  /* did not find any duplicated records  */
	else
		if (inkt == otkt)
			k = 0;  /* records read = records written, no dupes */
		else
			k = 1;  /* read more than written have dupes */
		
	return k;        
}

